import base64
import builtins
import inspect
import math

from .constants import PRIMITIVE_TYPES, UNSERIALIZABLE_DUNDER, \
    UNSERIALIZABLE_TYPES, UNSERIALIZABLE_CODE_TYPES, \
    ITERATOR_TYPE, CODE_TYPE, CELL_TYPE, MODULE_TYPE, FUNCTION_TYPE, \
    BYTES_TYPE, CLASS_TYPE, OBJ_TYPE, TUPLE_TYPE, SET_TYPE
from .supportive import get_fathers_class_for_method

from types import ModuleType, CellType, FunctionType, \
    MethodType, CodeType


class Converter:
    @classmethod
    def convert(cls, obj):
        if isinstance(obj, PRIMITIVE_TYPES):
            return obj
        if isinstance(obj, ModuleType):
            return cls._encode_module(obj)
        if isinstance(obj, CellType):
            return cls._encode_cell(obj)
        if isinstance(obj, bytes):
            return cls._encode_bytes(obj)
        if isinstance(obj, list):
            return type(obj)((cls.convert(item) for item in obj))
        if isinstance(obj, (tuple, set)):
            return cls._encode_collection(obj)
        if isinstance(obj, dict):
            return {key: cls.convert(value) for key, value in obj.items()}
        if isinstance(obj, (FunctionType, MethodType)):
            return cls._encode_function(obj)
        if isinstance(obj, type):
            return cls._encode_class(obj)
        if cls.is_iterable(obj):
            return cls._encode_iterator(obj)
        if isinstance(obj, CodeType):
            return cls._encode_code(obj)
        if isinstance(obj, object):
            return cls._encode_object(obj)

    @classmethod
    def deconvert(cls, obj):
        if isinstance(obj, PRIMITIVE_TYPES):
            return obj
        if isinstance(obj, list):
            return type(obj)((cls.deconvert(item) for item in obj))
        if isinstance(obj, dict):
            type_to_decode = cls._get_type(obj)

            if type_to_decode is None:
                return {key: cls.deconvert(value) for key, value in obj.items()}
            if type_to_decode == BYTES_TYPE:
                return cls._decode_bytes(obj)
            if type_to_decode == FUNCTION_TYPE:
                return cls._decode_function(obj)
            if type_to_decode == CELL_TYPE:
                return cls._decode_cell(obj)
            if type_to_decode == CLASS_TYPE:
                return cls._decode_class(obj)
            if type_to_decode == ITERATOR_TYPE:
                return cls._decode_iterator(obj)
            if type_to_decode == CODE_TYPE:
                return cls._decode_code(obj)
            if type_to_decode == MODULE_TYPE:
                return cls._decode_module(obj)
            if type_to_decode in (TUPLE_TYPE, SET_TYPE):
                return cls._decode_collection(obj)
            if type_to_decode == OBJ_TYPE:
                return cls._decode_object(obj)
        return obj

    @classmethod
    def _decode_collection(cls, obj):
        data = cls._get_data(obj)
        collection = getattr(builtins, cls._get_type(obj).lower())
        return collection((cls.deconvert(item) for item in data))

    @classmethod
    def _decode_module(cls, obj):
        return __import__(cls._get_data(obj))

    @classmethod
    def _decode_iterator(cls, obj):
        data = cls._get_data(obj)
        return iter(cls.deconvert(value) for value in data)

    @classmethod
    def _decode_code(cls, obj):
        data = cls._get_data(obj)

        def func():
            pass

        code_dict = cls.deconvert(data)
        return func.__code__.replace(**code_dict)


    @classmethod
    def _decode_object(cls, obj):
        data = cls._get_data(obj)
        obj_class = cls.deconvert(data['__class__'])

        decoded_obj = object.__new__(obj_class)
        decoded_obj.__dict__ = {
            key: cls.deconvert(value) for key, value in data['attrs'].items()
        }

        return decoded_obj

    @classmethod
    def _decode_class(cls, obj):
        data = cls._get_data(obj)

        class_bases = tuple(cls.deconvert(base) for base in data.pop('__bases__'))

        class_dict = {
            attr: cls.deconvert(value)
            for attr, value in data.items()
            if not (isinstance(value, dict) and cls._get_type(value) == FUNCTION_TYPE)
        }

        decoded_class = type(data['__name__'], class_bases, class_dict)

        for key, value in data.items():
            if isinstance(value, dict) and cls._get_type(value) == FUNCTION_TYPE:
                try:
                    function = cls.deconvert(value)
                except ValueError:
                    closure = cls._get_data(value)['closure']
                    cls._get_data(closure).append(cls._make_cell(decoded_class))
                    function = cls.deconvert(value)
                function.__globals__.update({decoded_class.__name__: decoded_class})

                if value.get('is_method'):
                    function = MethodType(function, decoded_class)

                setattr(decoded_class, key, function)
        return decoded_class

    @classmethod
    def _decode_cell(cls, obj):
        return cls._make_cell(cls.deconvert(cls._get_data(obj)))

    @classmethod
    def _decode_function(cls, obj):
        encoded_function = cls.deconvert(cls._get_data(obj))

        func_dict = encoded_function.pop('func_dict')

        new_func = FunctionType(**encoded_function)
        new_func.__dict__.update(func_dict)
        new_func.__globals__.update({new_func.__name__: new_func})
        return new_func

    @classmethod
    def _decode_bytes(cls, obj):
        return base64.b64decode(cls._get_data(obj).encode("ascii"))

    @classmethod
    def _encode_cell(cls, obj):
        data = cls.convert(obj.cell_contents)
        return cls._create_dict(data, CELL_TYPE)

    @classmethod
    def _encode_code(cls, obj):
        attrs = [attr for attr in dir(obj) if attr.startswith('co')]

        code_dict = {
            attr: cls.convert(getattr(obj, attr))
            for attr in attrs
            if attr not in UNSERIALIZABLE_CODE_TYPES
        }

        return cls._create_dict(code_dict, CODE_TYPE)

    @classmethod
    def _encode_iterator(cls, obj):
        data = list(map(cls.convert, obj))
        return cls._create_dict(data, ITERATOR_TYPE)

    @classmethod
    def _encode_class(cls, obj):
        data = {
            attr: cls.convert(getattr(obj, attr))
            for attr, value in inspect.getmembers(obj)
            if attr not in UNSERIALIZABLE_DUNDER
            and type(value) not in UNSERIALIZABLE_TYPES
        }

        data["__bases__"] = [
            cls.convert(base) for base in obj.__bases__ if base != object
        ]

        data["__name__"] = obj.__name__

        return cls._create_dict(data, CLASS_TYPE)

    @classmethod
    def _encode_object(cls, obj):
        data = {
            '__class__': cls.convert(obj.__class__),
            'attrs': {
                attr: cls.convert(value)
                for attr, value in inspect.getmembers(obj)
                if not attr.startswith('__')
                and not isinstance(value, FunctionType)
                and not isinstance(value, MethodType)
            }
        }
        return cls._create_dict(data, OBJ_TYPE)


    @classmethod
    def _encode_function(cls, obj):
        func_code = obj.__code__
        func_name = obj.__name__
        func_defaults = obj.__defaults__
        func_dict = obj.__dict__
        func_class = get_fathers_class_for_method(obj)
        func_closure = (
            tuple(cell for cell in obj.__closure__ if cell.cell_contents is not func_class)
            if obj.__closure__ is not None
            else tuple()
        )
        func_globs = {
            key: cls.convert(value)
            for key, value in obj.__globals__.items()
            if key in obj.__code__.co_names
            and value is not func_class
            and key != obj.__code__.co_name
        }

        encoded_func = cls.convert(
            dict(
                code=func_code,
                name=func_name,
                argdefs=func_defaults,
                closure=func_closure,
                func_dict=func_dict,
                globals=func_globs,
            )
        )
        return cls._create_dict(encoded_func, FUNCTION_TYPE, is_method=isinstance(obj, MethodType))

    @classmethod
    def _encode_collection(cls, obj):
        data = [cls.convert(item) for item in obj]
        return cls._create_dict(data, type(obj).__name__.lower())

    @classmethod
    def _encode_bytes(cls, obj: bytes):
        data = base64.b64encode(obj).decode("ascii")
        return cls._create_dict(data, BYTES_TYPE)

    @classmethod
    def _encode_module(cls, obj):
        return cls._create_dict(obj.__name__, MODULE_TYPE)

    @staticmethod
    def is_iterable(obj):
        return hasattr(obj, '__iter__') \
            and hasattr(obj, '__next__') \
            and callable(obj.__iter__)

    @staticmethod
    def _get_type(obj):
        if isinstance(obj, dict):
            return obj.get('__type')

    @staticmethod
    def _make_cell(value):
        return (lambda: value).__closure__[0]

    @staticmethod
    def _get_data(obj):
        if isinstance(obj, dict):
            return obj.get('data')

    @staticmethod
    def _create_dict(data, _type, **additional):
        return dict(__type=_type, data=data, **additional)



import types

JSON_DATA_TYPE = '.json'

XML_DATA_TYPE = '.xml'

TOML_DATA_TYPE = '.toml'

YAML_DATA_TYPE = '.yaml'

PRIMITIVE_TYPES = (int, str, bool, float, type(None))

UNKNOWN_TYPE_ERROR = 'Unknown type of serialization!'

SAME_TYPE_ERROR = 'Same type of objects!'

UNSUPPORTED_TYPE_ERROR = 'Unsupported data type: %s (%s)'

ITERATOR_TYPE = 'iterator'
BYTES_TYPE = 'bytes'
CODE_TYPE = 'code'
OBJ_TYPE = 'object'
MODULE_TYPE = 'module'
CELL_TYPE = 'cell'
FUNCTION_TYPE = 'function'
CLASS_TYPE = 'type'
SET_TYPE = 'set'
DICT_TYPE = 'dict'
TUPLE_TYPE = 'tuple'


UNSERIALIZABLE_DUNDER = (
    "__mro__",
    "__base__",
    "__basicsize__",
    "__class__",
    "__dictoffset__",
    "__name__",
    "__qualname__",
    "__text_signature__",
    "__itemsize__",
    "__flags__",
    "__weakrefoffset__",
    "__objclass__",
    "__doc__"
)

UNSERIALIZABLE_CODE_TYPES = (
    "co_positions",
    "co_lines",
    "co_exceptiontable",
    "co_lnotab",
)

UNSERIALIZABLE_TYPES = (
    types.WrapperDescriptorType,
    types.MethodDescriptorType,
    types.BuiltinFunctionType,
    types.MappingProxyType,
    types.GetSetDescriptorType,
)




from .constants import JSON_DATA_TYPE, TOML_DATA_TYPE, \
    YAML_DATA_TYPE, UNKNOWN_TYPE_ERROR, XML_DATA_TYPE
from .json_serializer import JsonSerializer
from .xml_serializer import XmlSerializer


class Factory:
    @staticmethod
    def create_serializer(data_format):
        if data_format == JSON_DATA_TYPE:
            return JsonSerializer()

        elif data_format == XML_DATA_TYPE:
            return XmlSerializer()

        else:
            raise Exception(UNKNOWN_TYPE_ERROR)






from .constants import UNKNOWN_TYPE_ERROR

from .converter import Converter


class JsonSerializer:
    _converter = Converter()

    @classmethod
    def load(cls, file):
        return cls.loads(file.read())

    @classmethod
    def loads(cls, obj):
        result, _ = cls._deconvert_from_string(obj)
        return cls._converter.deconvert(result)

    @classmethod
    def dump(cls, obj, file):
        file.write(cls.dumps(obj))

    @classmethod
    def dumps(cls, obj):
        return cls._convert_to_json_str(cls._converter.convert(obj))

    @classmethod
    def _convert_to_json_str(cls, obj):
        if isinstance(obj, bool):
            return str(obj).lower()
        if isinstance(obj, (int, float)):
            return str(obj)
        if isinstance(obj, str):
            return f'"{obj}"'
        if isinstance(obj, type(None)):
            return "null"
        if isinstance(obj, (list, set, tuple)):
            return f"[{','.join(list(map(cls.dumps, obj)))}]"
        if isinstance(obj, dict):
            data = ",".join(
                [f'"{key}": {cls.dumps(value)}' for (key, value) in obj.items()]
            )
            return f"""{{{data}}}"""
        raise Exception(UNKNOWN_TYPE_ERROR)

    @classmethod
    def _deconvert_from_string(cls, string_obj, start_index=0):
        if string_obj[start_index] == '"':
            return cls._loads_string(string_obj, start_index)
        elif string_obj[start_index] == '[':
            return cls._loads_list(string_obj, start_index)
        elif string_obj[start_index].isdigit() or string_obj[start_index] == '-':
            return cls._loads_nums(string_obj, start_index)
        elif string_obj[start_index] == 't':
            return True, start_index + 4
        elif string_obj[start_index] == 'f':
            return False, start_index + 5
        elif string_obj[start_index] == 'n':
            return None, start_index + 4
        elif string_obj[start_index] == '{':
            return cls._loads_dict(string_obj, start_index)
        raise Exception(UNKNOWN_TYPE_ERROR)

    @classmethod
    def _loads_nums(cls, obj, start_index):
        end_index = start_index + 1
        while len(obj) > end_index and (obj[end_index].isdigit() or obj[end_index] == '.'):
            end_index += 1

        num = obj[start_index:end_index]

        if num.count('.'):
            return float(num), end_index

        return int(num), end_index

    @classmethod
    def _loads_string(cls, obj, start_index):
        end_index = start_index + 1
        while obj[end_index] != '"':
            end_index += 1

        return obj[start_index + 1:end_index], end_index + 1

    @classmethod
    def _loads_list(cls, obj, start_index):
        end_index = start_index + 1
        braces = 1
        while braces:
            if obj[end_index] == '[':
                braces += 1
            if obj[end_index] == ']':
                braces -= 1

            end_index += 1

        output_list = []
        index = start_index + 1
        while index < end_index - 2:
            while obj[index] in (' ', ',', '\n'):
                index += 1
            res, index = cls._deconvert_from_string(obj, index)
            output_list.append(res)

        return output_list, end_index

    @classmethod
    def _loads_dict(cls, obj, start_index):
        end_index = start_index + 1
        braces = 1
        while braces:
            if obj[end_index] == '{':
                braces += 1
            if obj[end_index] == '}':
                braces -= 1

            end_index += 1

        index = start_index + 1
        output_dict = {}
        while index < end_index - 2:
            while obj[index] in (' ', ',', '\n'):
                index += 1
            key, index = cls._loads_string(obj, index)
            while obj[index] in (' ', ',', '\n', ':'):
                index += 1

            value, index = cls._deconvert_from_string(obj, index)
            output_dict[key] = value

        return output_dict, end_index + 1




import inspect


def get_fathers_class_for_method(method):
    cls = getattr(
        inspect.getmodule(method),
        method.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
        None
    )
    if isinstance(cls, type):
        return cls


def is_iterable(obj):
    return hasattr(obj, '__iter__') \
        and hasattr(obj, '__next__') \
        and callable(obj.__iter__)




from setuptools import setup

setup(
    name='tivan-custom-ser',
    version='0.1',
    packages=[
        "lab3.serializer.src"
    ],
    entry_points={
        'console_scripts': [
            "custom-serialize = lab3.serializer.custom:main"
        ]
    },
    license='MIT',
    author='tivan',
    author_email='tenugo.v@gmail.com',
    description='Python JSON and XML serializer',
)



from .constants import UNKNOWN_TYPE_ERROR
from .converter import Converter


class XmlSerializer:
    _converter = Converter()

    @classmethod
    def dumps(cls, obj):
        return cls._convert_to_xml_string(cls._converter.convert(obj))

    @classmethod
    def dump(cls, obj, file):
        file.write(cls.dumps(obj))

    @classmethod
    def loads(cls, obj):
        result, _ = cls._deconvert_from_xml_string(obj)
        return cls._converter.deconvert(result)

    @classmethod
    def load(cls, file):
        return cls.loads(file.read())

    @classmethod
    def _convert_to_xml_string(cls, obj):
        if isinstance(obj, bool):
            return f'<bool>{str(obj)}</bool>'
        if isinstance(obj, int):
            return f'<int>{str(obj)}</int>'
        if isinstance(obj, float):
            return f'<float>{str(obj)}</float>'
        if isinstance(obj, str):
            return f'<str>{obj}</str>'
        if isinstance(obj, type(None)):
            return f'<none>None</none>'
        if isinstance(obj, (list, tuple)):
            return f"<list>{''.join(list(map(cls.dumps, obj)))}</list>"
        if isinstance(obj, dict):
            data = ''.join(
                [f'<{key}>{cls.dumps(value)}</{key}>' for (key, value) in obj.items()]
            )
            return f'<dict>{data}</dict>'
        raise Exception(UNKNOWN_TYPE_ERROR)

    @classmethod
    def _deconvert_from_xml_string(cls, obj, start_position=0):
        index = start_position
        if obj[index] != '<':
            raise Exception(f'Invalid symbol at position {index}')

        type_start_index = index + 1
        type_end_index = index

        while obj[type_end_index] != '>':
            type_end_index += 1

        obj_type = obj[type_start_index:type_end_index]
        method_name = f'_loads_{obj_type}'

        if not hasattr(cls, method_name):
            raise Exception(UNKNOWN_TYPE_ERROR)
        index = type_end_index + 1

        return getattr(cls, method_name)(obj, index)

    @classmethod
    def _loads_str(cls, obj, start_position):
        end_position = start_position
        while obj[end_position:end_position + 6] != '</str>':
            end_position += 1

        return obj[start_position:end_position], end_position + 6

    @classmethod
    def _loads_bool(cls, obj, start_position):
        end_position = start_position
        while obj[end_position:end_position + 7] != '</bool>':
            end_position += 1

        bool_obj = obj[start_position:end_position]

        if bool_obj == 'True':
            return True, end_position + 7
        else:
            return False, end_position + 7

    @classmethod
    def _loads_int(cls, obj, start_position):
        end_position = start_position
        while obj[end_position:end_position + 6] != '</int>':
            end_position += 1

        int_obj = obj[start_position:end_position]
        return int(int_obj), end_position + 6

    @classmethod
    def _loads_float(cls, obj, start_position):
        end_position = start_position
        while obj[end_position:end_position + 8] != '</float>':
            end_position += 1

        float_obj = obj[start_position:end_position]
        return float(float_obj), end_position + 8

    @classmethod
    def _loads_none(cls, obj, start_position):
        end_position = start_position
        while obj[end_position:end_position + 7] != '</none>':
            end_position += 1

        return None, end_position + 7

    @classmethod
    def _loads_list(cls, obj, start_position):
        end_position = start_position
        deep = 1
        while deep:
            if obj[end_position:end_position + 6] == '<list>':
                deep += 1
            if obj[end_position:end_position + 7] == '</list>':
                deep -= 1

            end_position += 1

        end_position -= 1
        arr = []
        position = start_position
        while position < end_position:
            result, position = cls._deconvert_from_xml_string(obj, position)
            arr.append(result)

        return arr, end_position + 7

    @classmethod
    def _loads_dict(cls, obj, start_position):
        end_position = start_position
        deep = 1
        while deep:
            if obj[end_position:end_position + 6] == '<dict>':
                deep += 1
            if obj[end_position:end_position + 7] == '</dict>':
                deep -= 1
            end_position += 1

        end_position -= 1

        position = start_position
        result = {}

        while position < end_position:
            key_start_position = position + 1
            key_end_position = position + 1

            while obj[key_end_position] != '>':
                key_end_position += 1

            key = obj[key_start_position:key_end_position]

            value, position = cls._deconvert_from_xml_string(obj, key_end_position + 1)
            position += 3 + len(key)

            result[key] = value

        return result, end_position + 7
